import os
import base64

import cryptography.fernet as fernet
import cryptography.hazmat.primitives.ciphers as ciphers
from cryptography.hazmat.backends import default_backend


def aes_cipher(key, iv, mode=None):
    algo = ciphers.algorithms.AES(key)
    mode = mode or ciphers.modes.CBC(iv)
    return ciphers.Cipher(algo, mode, backend=default_backend())


def aes_encryptor(key, mode=None):
    iv = os.urandom(ciphers.algorithms.AES.block_size // 8)
    cipher = aes_cipher(key, iv, mode=mode)
    return cipher.encryptor(), iv


def aes_decryptor(key, iv, mode=None):
    cipher = aes_cipher(key, iv=iv, mode=mode)
    return cipher.decryptor()


def fernet_cipher(key):
    return fernet.Fernet(base64.urlsafe_b64encode(key))


def encrypt(data, key):
    """
    Encrypts binary data using cryptography's default fernet algorithm
    :param data: plaintext data to encrypt
    :param key: encryption key
    :return: encrypted data as a fernet token (a signed, base64 encoded string)
    """
    return fernet_cipher(key).encrypt(data)


def decrypt(data, key):
    """
    Decrypts binary data using cryptography's default fernet algorithm
    :param data: a fernet token to decrypt
    :param key: encryption key
    :return: decrypted data
    """
    return fernet_cipher(key).decrypt(data)


def encrypt_str(data, key):
    """
    Encrypts a unicode string using cryptography's default fernet algorithm
    :param data: unicode string to encrypt
    :param key: encryption key
    :return: encrypted utf-8 string as a fernet token
    """
    bin_data = data.encode('utf-8')
    return encrypt(bin_data, key)


def decrypt_str(cipher_text, key):
    """
    Decrypts a unicode string using cryptography's default fernet algorithm
    :param cipher_text: a utf-8 string as a fernet token to decrypt
    :param key: encryption key
    :return: decrypted unicode string
    """
    bin_data = decrypt(cipher_text, key)
    return bin_data.decode('utf-8')



import cryptography
import pytest
from blazeutils import randchars
from cryptography.hazmat.primitives import ciphers

from keg_elements import crypto


key = b'abcdefghijklmnopqrstuvwxyz123456'


def test_encrypt():
    data = b'\x01\x02\x03\x04\x05'
    encrypted = crypto.encrypt(data, key)
    assert data not in encrypted
    assert encrypted != crypto.encrypt(data, key)


def test_decrypt():
    data = b'\x01\x02\x03\x04\x05'
    assert crypto.decrypt(
        b'gAAAAABYVEWYMFW8Nyg91RlWubjVe_B82Q7AIyNDemSQSylUnJefKZ3k9smRdEP9-rD2vyNirrXqWCZ3wN1gk-Fu'
        b'gcdkwGuelQ==',
        key
    ) == data
    assert crypto.decrypt(
        b'gAAAAABYVEXjvHp-r0FnZYJ5t3_p1H9TDMU6lGOAAeHerYkHR40XHCfEJl1FEI0w6WsmwW5Gc1976LVyf46jX4K3'
        b'BPfWkY8aNA==',
        key
    ) == data

    with pytest.raises(cryptography.fernet.InvalidToken):
        crypto.decrypt(
            b'gAAAAABYVEXjvHp-r0FnZYJ5t3_p1H9TDMU6lGOAAeHerYkHR40XHCfEJl1FEI0w6WsmwW5Gc1976LVyf46j'
            b'X4K3BPfWkY8aNA==',
            b'a' * 32
        )


def test_encrypt_str():
    s = randchars()
    encrypted = crypto.encrypt_str(s, key)
    assert s.encode() not in encrypted
    assert encrypted != crypto.encrypt_str(s, key)


def test_decrypt_str():
    assert crypto.decrypt_str(
        b'gAAAAABYVEJVKXq6WlWn0wo3DDakIr6nwtHtk6llws4uJxgVvoBPLfhNRZc5eiWJEEAsEvtUjzQBqfaNscmqy2U'
        b'h0J21n1TyDA==',
        key
    ) == 'foo'
    assert crypto.decrypt_str(
        b'gAAAAABYVENwMqAohDnzfTAZy6P4SV1Z8fOkREMNd52Oey1SeVE5aZcgcJBomEVFORJjMvh5LIUEQ2vgeVI-bi'
        b'CDjboq37uOaw==',
        key
    ) == 'foo'


class TestCryptors:
    def enc_algo(self, encryptor):
        return encryptor._ctx._cipher.name, encryptor._ctx._mode.name

    def test_aes_encryptor_defaults(self):
        enc1, iv1 = crypto.aes_encryptor(key)
        enc2, iv2 = crypto.aes_encryptor(key)
        assert self.enc_algo(enc1) == ('AES', 'CBC')
        assert self.enc_algo(enc2) == ('AES', 'CBC')

        assert iv1 != iv2
        assert len(iv1) == 16
        assert len(iv2) == 16

    def test_aes_encryptor_mode(self):
        enc, iv = crypto.aes_encryptor(key, mode=ciphers.modes.ECB())
        assert self.enc_algo(enc) == ('AES', 'ECB')

    def test_aes_decryptor_defaults(self):
        dec1 = crypto.aes_decryptor(key, b'a' * 16)
        dec2 = crypto.aes_decryptor(key, b'a' * 16)
        assert self.enc_algo(dec1) == ('AES', 'CBC')
        assert self.enc_algo(dec2) == ('AES', 'CBC')

    def test_aes_decryptor_mode(self):
        dec = crypto.aes_decryptor(key, b'a' * 16, mode=ciphers.modes.ECB())
        assert self.enc_algo(dec) == ('AES', 'ECB')

    def test_encrypt_decrypt(self):
        data = b'a' * 32
        enc, iv = crypto.aes_encryptor(key)
        enc_data = enc.update(data) + enc.finalize()

        # check CBC is functioning
        assert enc_data[:16] != enc_data[16:]

        dec = crypto.aes_decryptor(key, iv)
        dec_data = dec.update(enc_data) + dec.finalize()
        assert dec_data == data
